/*

  PSID2BIN 20.03.2004 Tim Bscke - t.boescke@tu-harburg.de
  Quick Ansi-C hack, should compile on any platform.

  This programs reads .sid tunes in PSID format and converts
  them to binary, .prg or into an executable form for the Plus/4
  with SID-Card. 
  */

#include <stdio.h>
#include "musicplayer.h"

void movesid(unsigned char *dat, int length, int newbase,int fVerbose);


int main(int argc, char **argv) {

	char *filename;
	char filenamestub[320]; // lets hope the filename is no longer than 320 chars..
	char filenamefull[320];
	FILE *filein,*fileout;
	int i;
	unsigned char filedat[65536]; //lalala...
	int inlen;
	int dataoffset;
	int initadr;
	int	playadr;
	int	loadadr;
	int	binlen;
	int fReloc=0;
	int	fVerbose=0;
	int	fSaveBin=0;
	int fSavePrg=0;
	int	fSaveNfo=0;
	int fSaveExec=0;

	if (argc<2) {

		printf("PSID 2 BIN Converter v0.1 3/2004 Tim Boescke\n");
		printf("\nUsage: psid2bin <options> filein\n");
		printf("\nOptions:\n");
		printf(" -v\tVerbose (Show relocation info)\n");
		printf(" -b\tWrite binary\n");
		printf(" -p\tWrite .prg file (first two bytes are loadaddress)\n");
		printf(" -n\tWrite .nfo file with song information\n");
		printf(" -x\tWrite executable player for Plus/4. -r will be automatically assumed!\n");
		printf(" -r\tRelocate file to Plus/4 SID address ($fd40)\n");
		return 0;
	}
	
	// Parse arguments

	for (i=1; i<(argc-1); i++) {
		int j=1;
		char c;
		if (argv[i][0]=='-') {
			while(c=argv[i][j]) {
				switch (c) {
					case 'v': fVerbose=1; break;
					case 'r': fReloc^=1; break;
					case 'b': fSaveBin=1; break;
					case 'p': fSavePrg=1; break;
					case 'n': fSaveNfo=1; break;
					case 'x': fSaveExec=1;fReloc^=1; break;
				}
				j++;
			}
		}
	}

	filename=argv[argc-1];
	filein=fopen(filename,"rb");
	if (!filein) {printf("File '%s' not found..\n",filename); return -1;}
	inlen=fread(filedat,1,65536,filein);
	fclose(filein);

	dataoffset=(filedat[6]<<8)+filedat[7];
	loadadr   =(filedat[8]<<8)+filedat[9];
	if (!loadadr) {
		loadadr=filedat[dataoffset]+(filedat[dataoffset+1]<<8);
		dataoffset+=2;
	}

	binlen	  =inlen-dataoffset;
	initadr   =(filedat[10]<<8)+filedat[11];
	playadr   =(filedat[12]<<8)+filedat[13];

	printf("Name:\t\t%s\n",filedat+0x16);
	printf("Author:\t\t%s\n",filedat+0x36);
	printf("Copyright:\t%s\n",filedat+0x56);
	printf("Loadadr:\t$%.4X-$%.4X\nLength:\t\t$%.4X\nInitadr:\t$%.4X\nPlayadr:\t$%.4X\n",loadadr,loadadr+binlen,binlen,initadr,playadr);
	printf("Songs:\t\t%i\nStartsong:\t%i\n",filedat[15],filedat[17]);

	
	// Reloc

	if (fReloc) movesid(filedat+dataoffset,binlen,0xfd40,fVerbose);

	// build filename
	i=0;
	while ((filename[i])&&(filename[i]!='.')) filenamestub[i]=filename[i++];
	filenamestub[i]=0;

	// Save binary
	if (fSaveBin) {
		sprintf(filenamefull,"%s.bin",filenamestub);
		fileout=fopen(filenamefull,"wb");
		fwrite(&filedat[dataoffset],1,binlen,fileout);
		fclose(fileout);
	}

	// Save .prg
	if (fSavePrg) {
		sprintf(filenamefull,"%s.prg",filenamestub);
		fileout=fopen(filenamefull,"wb");
		fputc(loadadr&255,fileout);
		fputc((loadadr>>8)&255,fileout);
		fwrite(&filedat[dataoffset],1,binlen,fileout);
		fclose(fileout);
	}

	// Save .nfo
	if (fSaveNfo) {
		sprintf(filenamefull,"%s.nfo",filenamestub);
		fileout=fopen(filenamefull,"wb");
		fprintf(fileout,"Name:\t\t%s\n",filedat+0x16);
		fprintf(fileout,"Author:\t\t%s\n",filedat+0x36);
		fprintf(fileout,"Copyright:\t%s\n",filedat+0x56);
		fprintf(fileout,"Loadadr:\t$%.4X-$%.4X\nLength:\t\t$%.4X\nInitadr:\t$%.4X\nPlayadr:\t$%.4X\n",loadadr,loadadr+binlen,binlen,initadr,playadr);
		fprintf(fileout,"Songs:\t\t%i\nStartsong:\t%i\n",filedat[15],filedat[17]);

		fclose(fileout);
	}

	// Save .Exec
	if (fSaveExec) {
		sprintf(filenamefull,"%s_Player.prg",filenamestub);
		fileout=fopen(filenamefull,"wb");
		fwrite(musicplayer,1,sizeof(musicplayer),fileout);
		/*
			;		  !word $1000           ;d2 tunestart
			;tunelen	!word $1bb1           ;d4 tunelen
			;tuneinit  !word $1000            ;d6 tuneinit
			;tuneplay  !word $1003            ;d8 tuneplay
			;initvalue !byte 0                ;da initvalue
		*/
		fputc(loadadr&255,fileout); fputc((loadadr>>8)&255,fileout);
		fputc(binlen&255,fileout); fputc((binlen>>8)&255,fileout);
		fputc(initadr&255,fileout); fputc((initadr>>8)&255,fileout);
		fputc(playadr&255,fileout); fputc((playadr>>8)&255,fileout);
		fputc(filedat[17]-1,fileout);
		fwrite(&filedat[dataoffset],1,binlen,fileout);
		fclose(fileout);
	}

	printf("%s\n",filein); 
	return 0;
}

/*
 Relocate all $D400 based SID writes to a new base address
*/
void movesid(unsigned char *dat, int length, int newbase,int fVerbose) {

	unsigned char opcodes[]={0x8c,0x8d,0x8e,0x9d,0x99,0xAD,0xAE,0xAC,0xbd,0xb9,0xbe,0x00};  // stx,sta,sty,sta ,x, sta ,y
	char *mnemonics[]={" STY  "," STA  "," STX  "," STA,X"," STA,Y",
					   "!LDA  ","!LDX  ","!LDY  ","!LDA,X","!LDA,Y","!LDX,Y"};
	int i;

	if (fVerbose) printf("\nRelocating..\n");

	for (i=0; i<(length-2); i++) {
		unsigned char cur=dat[i];
		int j=0;

		do {
			if (cur==opcodes[j]) {
				int adr=dat[i+1]+dat[i+2]*256;

				if ((adr>=0xd400) && (adr<=0xd41f)) {
					int newadr=adr-0xd400+newbase;

					dat[i+1]=newadr&255;
					dat[i+2]=(newadr>>8)&255;

					if (fVerbose) printf("%.4X %s $%.4X -> $%.4X\n",i,mnemonics[j],adr,newadr);
				}

				i+=2;
				break;
			}
		} while (opcodes[j++]);	
	}
}
